    /******************************************************************
     editor.js                                              Muze Helene
     ------------------------------------------------------------------
     Author: Muze (info@muze.nl)
     Date: 28 februari 2004

     Copyright 2002 Muze

     This file is part of Helene.

     Helene is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published 
     by the Free Software Foundation; either version 2 of the License, 
     or (at your option) any later version.
 
     Helene is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with Helene; if not, write to the Free Software 
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
     02111-1307  USA

    -------------------------------------------------------------------

     This file contains the basic editor functionality.

    ******************************************************************/
	var currentLine=1;
	var lastLine=0;
	var screenWidth=0;
	var lineHeight=16;
	var contentDiv;
	var lines;
	var inputLine;
	var inputLineEntry;
	var oldpos=0;
	var isDirty=false;
	var winW = 630;
	var winH = 460;
	var inputLineHasFocus = false;
	var offsetLine = 0;

	function init() {
		if (parseInt(navigator.appVersion)>3) {
			if (navigator.appName=="Netscape") {
				winW = window.innerWidth-42;
				winH = window.innerHeight-2;
				lineHeight=18;
			}
			if (navigator.appName.indexOf("Microsoft")!=-1) {
				winW = document.body.offsetWidth;
				winH = document.body.offsetHeight;
			}
		}
		contentDiv=document.getElementById('content');	
		codeframeDiv=document.getElementById('codeframe');
		lines=document.getElementById('lines');
		lastLine=lines.childNodes.length;
		inputLine=document.getElementById('inputLine');
		inputLineEntry=document.getElementById('inputLineEntry');
		document.getElementById('linenumbers').style.height=contentDiv.offsetHeight;
		inputLine.style.width=contentDiv.clientWidth-44;
		inputLine.style.height=lineHeight;
		inputLineEntry.onfocus=inputFocus;
		inputLineEntry.onblur=inputBlur;
		document.onkeydown=KeyDownHandler;
		document.onkeyup=KeyUpHandler;
		document.onmousedown=MouseDownHandler;
		document.onmouseup=MouseUpHandler;
		var _temp='';
		for (var i=0; i<lines.childNodes.length; i++) {
			// need to remove trailing ' ', which IE puts there.
			_temp=getLineContent(i+1);
			if (navigator.appName.indexOf("Microsoft")!=-1) {
				_temp=_temp.substr(0, _temp.length-1);
			}
			currLine=new hLine(i, _temp);
			currLine.doHighlight(applyHighlightedLine);
		}
		redrawInputLine();
	}

	function inputFocus() {
		inputLineHasFocus=true;
	}

	function inputBlur() {
		inputLineHasFocus=false;
	}

	function applyHighlightedLine(lineNo, highlightedLine) {
		setLineContent(lineNo+1, '<nobr>'+highlightedLine+'</nobr>');
	}

	var currInputHeight=lineHeight;

	function moveDown(amount) {
		//FIXME: start made with selections
		//  change moveUp as well
		//  and change textarea size back again in stopselect
		if (userSelecting) {
			currInputHeight+=lineHeight;
			inputLine.style.height=currInputHeight;
			inputLineEntry.style.height=currInputHeight;
			offsetLine++;
			inputLineEntry.value+='\n'+getLineContent(currentLine+offsetLine);
			return true;
		} else {
			checkDirty();
			if (currentLine<lastLine) {
				currentLine+=amount;
				if (currentLine>lastLine) {
					currentLine=lastLine;
				}
				redrawInputLine();
			}
			return false;
		}
	}

	function moveUp(amount) {
		checkDirty();
		if (currentLine>1) {
			currentLine-=amount;
			if (currentLine<1) {
				currentLine=1;
			}
			redrawInputLine();
		}
	}

	function redrawInputLine() {

		var pos=getCursorPos();
		if (pos>oldpos) {
			oldpos=pos;
		}

		inputLine.style.top=(currentLine-1)*lineHeight-1;
		var value=getLineContent(currentLine);
		inputLineEntry.value=value;
		if (oldpos) {
			pos=oldpos;
		}
		setCursorPos(pos);
	}

	function getCursorPos() {
		if (!inputLineHasFocus) {
			inputLineEntry.focus()
		}
		if (document.selection) {
			var mtext=document.selection.createRange();
			var count=0;
			var moved=0;
			while (moved=mtext.moveStart('character', -100)) {
				count-=moved;
			}
			chars=mtext.text.length;
			return chars;
		} else {
			var selection=window.getSelection();
			if (selection.anchorNode) {
				// do something mozilla style
			}
		}
	}

	function setCursorPos(chars, clicked) {
		if (!inputLineHasFocus) {
			inputLineEntry.focus();
		}
		var contents=new String(inputLineEntry.value);
		if (document.selection) {
			var mtext=document.selection.createRange();
			if (contents.length<chars) {
				chars=contents.length;
			}
			// it appears that IE doesn't like it if the range
			// leaves the textarea even one character, so be 
			// extra carefull to move exactly to the start
			// default cursor location is exactly at the end
			var mchars=contents.length-chars;
			var curpos=getCursorPos();
			mchars=chars-curpos;
			mtext.moveStart('character', mchars);
			mtext.collapse(true);
			mtext.select();
		} else {
			var selection=window.getSelection();
			if (selection.anchorNode) {
				// do something mozilla style
			}
		}
//		inputLineEntry.focus();
		return false;
	}

	function getLineContent(lineNo) {
		if (lines.childNodes[lineNo-1].innerText) {
			return lines.childNodes[lineNo-1].innerText;
		} else if (document.createRange) {
			var html = document.createRange();
			var myLine = lineNo-1;
			while( lines.childNodes[myLine].tagName != 'LI' ) {
				myLine++;
			}
			html.selectNodeContents(lines.childNodes[myLine]);
			var myString = html.toString();
			myString = myString.replace(/\n/g, '');
			return myString;
		}
		return "";
	}

	function KeyDownHandler(evt) {
		evt = (evt) ? evt : ((event) ? event : null );
		if( evt ) {
			var charCode = (evt.charCode ? evt.charCode : ((evt.keyCode) ? evt.keyCode : evt.which));
			if( charCode==40 ) { // cursor down
				if (evt.shiftKey) {
					startSelect();
				} else {
					stopSelect();
				}
				return moveDown(1);
			} else if (charCode==38) { // cursor up
				if (evt.shiftKey) {
					startSelect();
				} else {
					stopSelect();
				}
				moveUp(1);
				return false;
			} else if (charCode==37) { // cursor left
				if (evt.shiftKey) {
					startSelect();
				} else {
					stopSelect();
				}
				var result=true;
				if (getCursorPos()==0) {
					if (currentLine>1) {
						moveUp(1);
						setCursorPos(999);
						result=false;
					}
				}
				oldpos=0;
				return result;
			} else if (charCode==39) { // cursor right
				if (evt.shiftKey) {
					startSelect();
				} else {
					stopSelect();
				}
				var result=true;
				if (currentLine<lastLine) {
					var _currline=getLineContent(currentLine);
					if (getCursorPos()==_currline.length) {
						moveDown(1);
						setCursorPos(0);
						result=false;
					}
				}
				oldpos=0;
				return result;
			} else if (charCode==8) { // backspace
				oldpos=0;
				if (getCursorPos()==0) {
					checkDirty();
					glueCurrentLine(-1);
					return false;
				}
				isDirty=true;
			} else if (charCode==46) { // delete
				oldpos=0;
				if (currentLine<lastLine) {
					var _editline=inputLineEntry.value;
					if (getCursorPos()==_editline.length) {
						checkDirty();
						glueCurrentLine(1);
						return false;
					}
				}
				isDirty=true;
			} else if (charCode==13) { // enter
				oldpos=0;
				breakCurrentLine();
				return false;
			} else if (charCode==34) { // pgdown
				// 20 lines down
				if (evt.shiftKey) {
					startSelect();
				} else {
					stopSelect();
				}
				moveDown(20);
				return false;
			} else if (charCode==33) { // pgup
				// 20 lines up
				if (evt.shiftKey) {
					startSelect();
				} else {
					stopSelect();
				}
				moveUp(20);
				return false;
			} else if ((charCode==86) && evt.ctrlKey) { // ctrl-v
//				alert('paste');
//				return false;
			} else if ((charCode==67) && evt.ctrlKey) { // ctrl-c
//				alert('copy');
//				return false;
			} else if ((charCode==88) && evt.ctrlKey) { // ctrl-x
//				alert('cut');
//				return false;
			} else if((charCode==65) && evt.ctrlKey) {
				setCursorPos(0);
				return false;
			} else {
				oldpos=0;
				isDirty=true;
				window.status=charCode;
			}
		}
	}

	function KeyUpHandler() {
	}

	function MouseUpHandler(evt) {
		evt = (evt) ? evt : event;
		var target = (evt.target) ? evt.target : evt.srcElement;
		// get click position
		if (evt.pageX) {
			var offsetX=evt.pageX - ((target.offsetLeft) ? target.offsetLeft : target.left);
			var offsetY=evt.pageY - ((target.offsetTop) ? target.offsetTop : target.top);				
		} else if (evt.offsetX || evt.offsetY) {
			var offsetX = evt.offsetX;
			var offsetY = evt.offsetY;	
		} else if (evt.clientX || evt.clientY) {
			var offsetX = evt.clientX - ((target.offsetLeft) ? target.offsetLeft : 0);
			var offsetY = evt.clientY - ((target.offsetTop) ? target.offsetTop : 0);
		}
		offsetX-=44;
		var offsetXchars=Math.round(offsetX/9);
		// find a known container of the mouseclick
		while (target.nodeName!='LI' && target.nodeName!='TEXTAREA' && target.nodeName!='DIV') {
			target=target.parentElement;
		}
		if (target.nodeName=='LI') {
			var count=1;
			var currLine=target;
			while (currLine=currLine.previousSibling) {
				count++;
			}
			currentLine=count;
			redrawInputLine();
//			alert('sp:'+offsetXchars);
			setCursorPos(offsetXchars, true);
//			alert('done');
			evt.cancelBubble =true;
			return false;
		} else if (target.nodeName=='INPUT') {
		} else if (target.nodeName=='TEXTAREA') {
		} else {
			currentLine=Math.round(offsetY/lineHeight)+1;
			redrawInputLine();
//			alert('sp2:'+offsetXchars);
			setCursorPos(offsetXchars, true);
//			alert('done');
			evt.cancelBubble = true;
			return false;
		}
	}

	function MouseDownHandler(evt) {
		evt = (evt) ? evt : event;
	}

	function startSelect() {
		userSelecting=true;
	}

	function stopSelect() {
		userSelecting=false;
		// get selected contents, put them in a buffer or something
	}

	function updateLine(lineNo, lineContent) {
		highlightUpdateLine(lineNo-1, lineContent, applyHighlightedLine);
	}

	function setLineContent(lineNo, lineContent) {
		lines.childNodes[lineNo-1].innerHTML=lineContent;
	}

	function insertLineAfter(lineNo, lineContent) {
		lines.childNodes[lineNo-1].insertAdjacentHTML("AfterEnd","<li>");
		highlightAppendLine(lineNo-1, lineContent, false); // do not display this yet
		currentLine=lineNo+1;
		updateLine(currentLine, lineContent);
		oldpos=0;
		redrawInputLine();
		document.getElementById('linenumbers').style.height=contentDiv.clientHeight;
		lastLine++;
	}

	function deleteLine(lineNo) {
		lines.childNodes[lineNo-1].removeNode(true);
		highlightDeleteLine(lineNo-1, applyHighlightedLine);
		document.getElementById('linenumbers').style.height=contentDiv.clientHeight;
		lastLine--;
	}

	function checkDirty() {
		if (isDirty) {
			var newContents=inputLineEntry.value;
			updateLine(currentLine, newContents);
			isDirty=false;
		} else {
		}
	}

	function getLineHead() {
		var _line=new String(getLineContent(currentLine));
		var _pos=getCursorPos();
		return _line.substr(0, _pos);
	}

	function getLineTail() {
		var _line=new String(getLineContent(currentLine));
		var _pos=getCursorPos();
		return _line.substr(_pos);
	}

	function breakCurrentLine() {
		checkDirty();
		var head=getLineHead();
		var tail=getLineTail();
		insertLineAfter(currentLine, tail);
		updateLine(currentLine-1, head);
		setCursorPos(0);
	}

	function glueCurrentLine(direction) {
		checkDirty();
		if (direction==-1) { 
			if (currentLine>1) {
				var tail=getLineContent(currentLine);
				var head=getLineContent(currentLine-1);
				updateLine(currentLine-1, head+tail);
				deleteLine(currentLine);
				currentLine--;
				redrawInputLine();
				setCursorPos(head.length);
			}			
		} else {
			if (currentLine<lastLine) {
				var tail=getLineContent(currentLine+1);
				var head=getLineContent(currentLine);
				updateLine(currentLine, head+tail);
				deleteLine(currentLine+1);
				redrawInputLine();
			}
		}
	}
